import warnings
from unittest import mock
from unittest.mock import Mock

from typing_extensions import Optional

from posit.connect.resources import (
    BaseResource,
    _contains_dict_key_values,
    _matches_exact,
    _matches_pattern,
)

config = Mock()
session = Mock()


class FakeResource(BaseResource):
    @property
    def foo(self) -> Optional[str]:
        return self.get("foo")


class TestResource:
    def test_init(self):
        p = mock.Mock()
        k = "foo"
        v = "bar"
        d = {k: v}
        r = FakeResource(p, **d)
        assert r._ctx == p

    def test__getitem__(self):
        warnings.filterwarnings("ignore", category=FutureWarning)
        k = "foo"
        v = "bar"
        d = {k: v}
        r = FakeResource(mock.Mock(), **d)
        assert r.__getitem__(k) == d.__getitem__(k)
        assert r[k] == d[k]

    def test__setitem__(self):
        warnings.filterwarnings("ignore", category=FutureWarning)
        k = "foo"
        v1 = "bar"
        v2 = "baz"
        d = {k: v1}
        r = FakeResource(mock.Mock(), **d)
        assert r[k] == v1
        r[k] = v2
        assert r[k] == v2

    def test__delitem__(self):
        warnings.filterwarnings("ignore", category=FutureWarning)
        k = "foo"
        v = "bar"
        d = {k: v}
        r = FakeResource(mock.Mock(), **d)
        assert k in r
        assert r[k] == v
        del r[k]
        assert k not in r

    def test_foo(self):
        k = "foo"
        v = "bar"
        d = {k: v}
        r = FakeResource(mock.Mock(), **d)
        assert r.foo == v


class TestContainsDictKeyValues:
    def test_empty_value_dict(self):
        r = FakeResource(mock.Mock(), foo={"a": 1, "b": 2})
        assert _contains_dict_key_values(r, "foo", {}) is True

    def test_matching_single_key_value(self):
        r = FakeResource(mock.Mock(), foo={"a": 1, "b": 2})
        assert _contains_dict_key_values(r, "foo", {"a": 1}) is True

    def test_matching_multiple_key_values(self):
        r = FakeResource(mock.Mock(), foo={"a": 1, "b": 2, "c": 3})
        assert _contains_dict_key_values(r, "foo", {"a": 1, "b": 2}) is True

    def test_non_matching_key_value(self):
        r = FakeResource(mock.Mock(), foo={"a": 1, "b": 2})
        assert _contains_dict_key_values(r, "foo", {"a": 2}) is False

    def test_missing_key_in_item(self):
        r = FakeResource(mock.Mock(), foo={"a": 1})
        assert _contains_dict_key_values(r, "foo", {"b": 2}) is False

    def test_missing_field_in_resource(self):
        r = FakeResource(mock.Mock())
        assert _contains_dict_key_values(r, "nonexistent", {"a": 1}) is False

    def test_non_dict_field_value(self):
        r = FakeResource(mock.Mock(), foo="not_a_dict")
        assert _contains_dict_key_values(r, "foo", {"a": 1}) is False

    def test_nested_dict_values(self):
        r = FakeResource(mock.Mock(), foo={"nested": {"x": 10}, "simple": "value"})
        assert _contains_dict_key_values(r, "foo", {"nested": {"x": 10}}) is True
        assert _contains_dict_key_values(r, "foo", {"nested": {"x": 20}}) is False


class TestMatchesPattern:
    def test_pattern_matches(self):
        r = FakeResource(mock.Mock(), name="test-app-123")
        assert _matches_pattern(r, "name", r"test-.*-\d+") is True

    def test_pattern_does_not_match(self):
        r = FakeResource(mock.Mock(), name="production-app")
        assert _matches_pattern(r, "name", r"test-.*-\d+") is False

    def test_missing_field_returns_false(self):
        r = FakeResource(mock.Mock())
        assert _matches_pattern(r, "nonexistent", ".*") is False

    def test_empty_pattern_matches_any_string(self):
        r = FakeResource(mock.Mock(), description="any text here")
        assert _matches_pattern(r, "description", "") is True

    def test_partial_match_returns_true(self):
        r = FakeResource(mock.Mock(), title="My Great App")
        assert _matches_pattern(r, "title", "Great") is True


class TestMatchesExact:
    def test_exact_match_returns_true(self):
        r = FakeResource(mock.Mock(), name="test-app")
        assert _matches_exact(r, "name", "test-app") is True

    def test_no_match_returns_false(self):
        r = FakeResource(mock.Mock(), name="test-app")
        assert _matches_exact(r, "name", "other-app") is False

    def test_missing_field_returns_false(self):
        r = FakeResource(mock.Mock())
        assert _matches_exact(r, "nonexistent", "value") is False



import responses
from responses import matchers

from posit.connect.client import Client
from posit.connect.oauth import types

from ..api import load_mock, load_mock_dict


class TestIntegrationDelete:
    @responses.activate
    def test(self):
        guid = "22644575-a27b-4118-ad06-e24459b05126"

        # behavior
        responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )

        mock_delete = responses.delete(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integration = c.oauth.integrations.get(guid)

        # invoke
        integration.delete()

        # assert
        assert mock_delete.call_count == 1


class TestIntegrationUpdate:
    @responses.activate
    def test(self):
        # data
        guid = "22644575-a27b-4118-ad06-e24459b05126"
        responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )

        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integration = c.oauth.integrations.get(guid)
        assert integration["guid"] == guid

        new_name = "New Name"

        fake_integration = load_mock_dict(f"v1/oauth/integrations/{guid}.json")
        fake_integration.update(name=new_name)
        assert fake_integration["name"] == new_name

        mock_update = responses.patch(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=fake_integration,
        )

        integration.update(name=new_name)
        assert mock_update.call_count == 1
        assert integration["name"] == new_name


class TestIntegrationsCreate:
    @responses.activate
    def test(self):
        # data
        guid = "22644575-a27b-4118-ad06-e24459b05126"
        fake_integration = load_mock_dict(f"v1/oauth/integrations/{guid}.json")

        # behavior
        mock_create = responses.post(
            "https://connect.example/__api__/v1/oauth/integrations",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
            match=[
                matchers.json_params_matcher(
                    {
                        "name": fake_integration["name"],
                        "description": fake_integration["description"],
                        "template": fake_integration["template"],
                        "config": fake_integration["config"],
                    },
                ),
            ],
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        integration = c.oauth.integrations.create(
            name=fake_integration["name"],
            description=fake_integration["description"],
            template=fake_integration["template"],
            config=fake_integration["config"],
        )

        # assert
        assert mock_create.call_count == 1
        assert integration["name"] == fake_integration["name"]
        assert integration["description"] == fake_integration["description"]
        assert integration["template"] == fake_integration["template"]
        assert integration["config"] == fake_integration["config"]


class TestIntegrationsFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = responses.get(
            "https://connect.example/__api__/v1/oauth/integrations",
            json=load_mock("v1/oauth/integrations.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None

        # invoke
        integrations = c.oauth.integrations.find()

        # assert
        assert mock_get.call_count == 1
        assert len(integrations) == 3
        assert integrations[0]["id"] == "3"
        assert integrations[1]["id"] == "4"
        assert integrations[2]["id"] == "5"


class TestIntegrationsGet:
    @responses.activate
    def test(self):
        guid = "22644575-a27b-4118-ad06-e24459b05126"

        # behavior
        mock_get = responses.get(
            f"https://connect.example/__api__/v1/oauth/integrations/{guid}",
            json=load_mock(f"v1/oauth/integrations/{guid}.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integration = c.oauth.integrations.get(guid)

        assert mock_get.call_count == 1
        assert integration["guid"] == guid


class TestIntegrationsFindBy:
    @responses.activate
    def test(self):
        # behavior
        responses.get(
            "https://connect.example/__api__/v1/oauth/integrations",
            json=load_mock("v1/oauth/integrations.json"),
        )

        # setup
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        integrations = c.oauth.integrations

        # invoke and assert
        # by guid
        found = integrations.find_by(guid="22644575-a27b-4118-ad06-e24459b05126")
        assert found is not None
        assert found["guid"] == "22644575-a27b-4118-ad06-e24459b05126"

        # by name (exact match)
        found = integrations.find_by(name="^keycloak integration$")
        assert found is not None
        assert found["name"] == "keycloak integration"

        # by name (partial match) - should find first match
        found = integrations.find_by(name="keycloak")
        assert found is not None
        assert found["name"] == "keycloak integration"

        # by description (exact match)
        found = integrations.find_by(description="^integration description$")
        assert found is not None
        assert found["description"] == "integration description"

        # by description (partial match)
        found = integrations.find_by(description="another")
        assert found is not None
        assert found["description"] == "another integration description"

        # by integration_type
        found = integrations.find_by(integration_type=types.OAuthIntegrationType.CUSTOM)
        assert found is not None
        assert found["template"] == "custom"
        assert found["name"] == "keycloak integration"  # first one

        found = integrations.find_by(integration_type="custom")
        assert found is not None
        assert found["template"] == "custom"
        assert found["name"] == "keycloak integration"  # first one

        # by auth_type (no match)
        found = integrations.find_by(auth_type=types.OAuthIntegrationAuthType.VIEWER)
        assert found is None

        found = integrations.find_by(auth_type="Viewer")
        assert found is None

        # by config
        found = integrations.find_by(config={"client_id": "rsconnect-oidc"})
        assert found is not None
        assert found["config"]["client_id"] == "rsconnect-oidc"
        assert found["name"] == "keycloak integration"  # first one

        # by config to find second integration
        found = integrations.find_by(config={"token_endpoint_auth_method": "client_secret_post"})
        assert found is not None
        assert found["config"]["token_endpoint_auth_method"] == "client_secret_post"
        assert found["name"] == "keycloak-post"

        # by multiple criteria
        found = integrations.find_by(integration_type="custom", name="keycloak-post")
        assert found is not None
        assert found["template"] == "custom"
        assert found["name"] == "keycloak-post"

        # find connect integration
        found = integrations.find_by(integration_type=types.OAuthIntegrationType.CONNECT)
        assert found is not None
        assert found["template"] == "connect"
        assert found["name"] == "Connect Visitor API Key"

        found = integrations.find_by(integration_type="connect")
        assert found is not None
        assert found["template"] == "connect"
        assert found["name"] == "Connect Visitor API Key"

        found = integrations.find_by(name="Connect Visitor API Key")
        assert found is not None
        assert found["name"] == "Connect Visitor API Key"

        found = integrations.find_by(description="Allows access to Connect APIs")
        assert found is not None
        assert found["description"] == "Allows access to Connect APIs on behalf of a Connect user."

        found = integrations.find_by(config={"scopes": "admin"})
        assert found is not None
        assert found["config"]["scopes"] == "admin"

        # no match
        found = integrations.find_by(name="Non-existent Integration")
        assert found is None



"""OAuth integration resources."""

from __future__ import annotations

from functools import partial

from typing_extensions import TYPE_CHECKING, List, Optional, overload

from ..resources import (
    BaseResource,
    Resources,
    _contains_dict_key_values,
    _matches_exact,
    _matches_pattern,
)
from .associations import IntegrationAssociations

if TYPE_CHECKING:
    from ..oauth import types


class Integration(BaseResource):
    """OAuth integration resource."""

    @property
    def associations(self) -> IntegrationAssociations:
        return IntegrationAssociations(self._ctx, integration_guid=self["guid"])

    def delete(self) -> None:
        """Delete the OAuth integration."""
        path = f"v1/oauth/integrations/{self['guid']}"
        self._ctx.client.delete(path)

    @overload
    def update(
        self,
        *args,
        name: str = ...,
        description: str = ...,
        config: dict = ...,
        **kwargs,
    ) -> None:
        """Update the OAuth integration.

        Parameters
        ----------
        name: str, optional
        description: str, optional
        config: dict, optional
        """

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the OAuth integration."""

    def update(self, *args, **kwargs) -> None:
        """Update the OAuth integration."""
        body = dict(*args, **kwargs)
        path = f"v1/oauth/integrations/{self['guid']}"
        response = self._ctx.client.patch(path, json=body)
        super().update(**response.json())


class Integrations(Resources):
    """Integrations resource."""

    @overload
    def create(
        self,
        *,
        name: str,
        description: Optional[str],
        template: str,
        config: dict,
    ) -> Integration:
        """Create an OAuth integration.

        Parameters
        ----------
        name : str
        description : Optional[str]
        template : str
        config : dict

        Returns
        -------
        Integration
        """

    @overload
    def create(self, **kwargs) -> Integration:
        """Create an OAuth integration.

        Returns
        -------
        Integration
        """

    def create(self, **kwargs) -> Integration:
        """Create an OAuth integration.

        Parameters
        ----------
        name : str
        description : Optional[str]
        template : str
        config : dict

        Returns
        -------
        Integration
        """
        path = "v1/oauth/integrations"
        response = self._ctx.client.post(path, json=kwargs)
        return Integration(self._ctx, **response.json())

    def find(self) -> List[Integration]:
        """Find OAuth integrations.

        Returns
        -------
        List[Integration]
        """
        path = "v1/oauth/integrations"
        response = self._ctx.client.get(path)
        return [
            Integration(
                self._ctx,
                **result,
            )
            for result in response.json()
        ]

    def find_by(
        self,
        integration_type: Optional[types.OAuthIntegrationType | str] = None,
        auth_type: Optional[types.OAuthIntegrationAuthType | str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        guid: Optional[str] = None,
        config: Optional[dict] = None,
    ) -> Integration | None:
        """Find an OAuth integration by various criteria.

        Parameters
        ----------
        integration_type : Optional[types.OAuthIntegrationType | str]
            The type of the integration (e.g., "aws", "azure").
        auth_type : Optional[types.OAuthIntegrationAuthType | str]
            The authentication type of the integration (e.g., "Viewer", "Service Account").
        name : Optional[str]
            A regex pattern to match the integration name. For exact matches, use `^` and `$`. For example,
            `^My Integration$` will match only "My Integration".
        description : Optional[str]
            A regex pattern to match the integration description. For exact matches, use `^` and `$`. For example,
            `^My Integration Description$` will match only "My Integration Description".
        guid : Optional[str]
            The unique identifier of the integration.
        config : Optional[dict]
            A dictionary of configuration key-value pairs to match against the integration's config. This will
            vary based on the integration type.

        Returns
        -------
        Integration | None
            The first matching integration, or None if no match is found.
        """
        filters = []
        if integration_type is not None:
            filters.append(partial(_matches_exact, key="template", value=integration_type))
        if auth_type is not None:
            filters.append(partial(_matches_exact, key="auth_type", value=auth_type))
        if name is not None:
            filters.append(partial(_matches_pattern, key="name", pattern=name))
        if description is not None:
            filters.append(partial(_matches_pattern, key="description", pattern=description))
        if guid is not None:
            filters.append(partial(_matches_exact, key="guid", value=guid))
        if config is not None:
            filters.append(partial(_contains_dict_key_values, key="config", value=config))

        for integration in self.find():
            if all(f(integration) for f in filters):
                return integration

        return None

    def get(self, guid: str) -> Integration:
        """Get an OAuth integration.

        Parameters
        ----------
        guid: str

        Returns
        -------
        Integration
        """
        path = f"v1/oauth/integrations/{guid}"
        response = self._ctx.client.get(path)
        return Integration(self._ctx, **response.json())



"""Client connection for Posit Connect."""

from __future__ import annotations

from typing_extensions import TYPE_CHECKING, Optional, overload

from . import hooks, me
from .auth import Auth
from .config import Config
from .content import Content
from .context import Context, ContextManager, requires
from .groups import Groups
from .metrics.metrics import Metrics
from .oauth.oauth import OAuth
from .oauth.types import OAuthIntegrationType, OAuthTokenType
from .resources import _PaginatedResourceSequence, _ResourceSequence
from .sessions import Session
from .system import System
from .tags import Tags
from .tasks import Tasks
from .users import User, Users
from .vanities import Vanities

if TYPE_CHECKING:
    from requests import Response

    from .environments import Environments
    from .packages import Packages


class Client(ContextManager):
    """
    Client connection for Posit Connect.

    This class provides an interface to interact with the Posit Connect API,
    allowing for authentication, resource management, and data retrieval.

    Parameters
    ----------
    api_key : str, optional
        API key for authentication
    url : str, optional
        Sever API URL

    Attributes
    ----------
    content: Content
        Content resource.
    environments: Environments
        Environments resource.
    groups: Groups
        Groups resource.
    me: User
        Current user resource.
    metrics: Metrics
        Metrics resource.
    oauth: OAuth
        OAuth resource.
    packages: Packages
        Packages resource.
    system: System
        System resource.
    tags: Tags
        Tags resource.
    tasks: Tasks
        Tasks resource.
    users: Users
        Users resource.
    vanities: Vanities
        Vanities resource.
    version: str
        The server version.
    """

    @overload
    def __init__(self) -> None:
        """Initialize a Client instance.

        Creates a client instance using credentials read from the environment.

        Environment Variables
        ---------------------
        CONNECT_SERVER - The Connect server URL.
        CONNECT_API_KEY - The API key credential for client authentication.

        Examples
        --------
        Client()
        """

    @overload
    def __init__(self, url: str) -> None:
        """Initialize a Client instance.

        Creates a client instance using a provided URL and API key credential read from the environment.

        Environment Variables
        ---------------------
        CONNECT_API_KEY - The API key credential for client authentication.

        Parameters
        ----------
        url : str
            The Connect server URL.

        Examples
        --------
        Client("https://connect.example.com)
        """

    @overload
    def __init__(self, url: str, api_key: str) -> None:
        """Initialize a Client instance.

        Parameters
        ----------
        url : str
            The Connect server URL.
        api_key : str
            The API key credential for client authentication.

        Examples
        --------
        >>> Client("https://connect.example.com", os.getenv("CONNECT_API_KEY"))
        """

    def __init__(self, *args, **kwargs) -> None:
        """Initialize a Client instance.

        Environment Variables
        ---------------------
        CONNECT_SERVER - The Connect server URL.
        CONNECT_API_KEY - The API key credential for client authentication.

        Parameters
        ----------
        *args
            Variable length argument list. Can accept:
            - (url: str)
                url: str
                    The Connect server URL.
            - (url: str, api_key: str)
                url: str
                    The Connect server URL.
                api_key: str
                    The API key credential for client authentication.

        **kwargs
            Keyword arguments. Can include 'url' and 'api_key'.

        Examples
        --------
        >>> Client()
        >>> Client("https://connect.example.com")
        >>> Client("https://connect.example.com", os.getenv("CONNECT_API_KEY"))
        >>> Client(api_key=os.getenv("CONNECT_API_KEY"), url="https://connect.example.com")
        """
        api_key = None
        url = None
        if len(args) == 1 and isinstance(args[0], str):
            url = args[0]
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            url = args[0]
            api_key = args[1]
        else:
            if "api_key" in kwargs and isinstance(kwargs["api_key"], str):
                api_key = kwargs["api_key"]
            if "url" in kwargs and isinstance(kwargs["url"], str):
                url = kwargs["url"]

        self.cfg = Config(api_key=api_key, url=url)
        session = Session()
        session.auth = Auth(config=self.cfg)
        session.hooks["response"].append(hooks.check_for_deprecation_header)
        session.hooks["response"].append(hooks.handle_errors)
        self.session = session
        self._ctx = Context(self)

    @requires("2025.01.0")
    def with_user_session_token(
        self,
        token: str,
        audience: Optional[str] = None,
    ) -> Client:
        """Create a new Client scoped to the user specified in the user session token.

        Create a new Client instance from a user session token exchange for an api key scoped to the
        user specified in the token (the user viewing your app). If running your application locally,
        a user session token will not exist, which will cause this method to result in an error needing
        to be handled in your application.

        Depending on the type of application you are building, the user session token is retrieved in
        a variety of ways. For example, in Streamlit and Shiny applications, the token is stored in the
        context or session object headers using the `Posit-Connect-User-Session-Token` key. For API
        applications, the token is added to the request headers.

        Parameters
        ----------
        token : str
            The user session token.
        audience : str, optional
            The audience for the token exchange. This is the integration GUID of the Connect API integration
            that is associate with the content. If not provided when there are multiple integrations, the
            function will attempt to determine the audience from the current content associations.

        Returns
        -------
        Client
            A new Client instance authenticated with an API key exchanged for the user session token.

        Raises
        ------
        ValueError
            If the provided token is `None` or empty or if the exchange response is malformed.
        ClientError
            If the token exchange request with the Connect Server fails.

        Examples
        --------
        ```python
        from posit.connect import Client

        client = Client().with_user_session_token("my-user-session-token")
        ```

        Example using user session token from Shiny session:
        ```python
        from posit.connect import Client
        from shiny.express import render, session

        client = Client()


        @reactive.calc
        def visitor_client():
            ## read the user session token and generate a new client
            user_session_token = session.http_conn.headers.get("Posit-Connect-User-Session-Token")
            return client.with_user_session_token(user_session_token)


        @render.text
        def user_profile():
            # fetch the viewer's profile information
            return visitor_client().me
        ```

        Example of when the visitor's token could not be retrieved (for
        example, if this app allows unauthenticated access) and handle
        that in cases where a token is expected.
        ```python
        from posit.connect import Client
        import requests

        # Simulate request without header
        mock_request = requests.Request()
        visitor_client = None
        token = request.headers.get("Posit-Connect-User-Session-Token")
        if token:
            visitor_client = Client().with_user_session_token(token)
        else:
            print("This app requires a user session token to operate.")
        ```
        """
        if token is None or token == "":
            raise ValueError("token must be set to non-empty string.")

        # If the audience is not provided and there are multiple associations,
        # we will try to find the Connect API integration GUID from the content resource.
        current_content_associations = self.content.get().oauth.associations.find()
        if audience is None and len(current_content_associations) > 1:
            connect_api_integration_guids = [
                a["oauth_integration_guid"]
                for a in current_content_associations
                if a.get("oauth_integration_template") == OAuthIntegrationType.CONNECT
            ]
            if len(connect_api_integration_guids) == 1:
                audience = connect_api_integration_guids[0]

        visitor_credentials = self.oauth.get_credentials(
            token,
            requested_token_type=OAuthTokenType.API_KEY,
            audience=audience,
        )

        visitor_api_key = visitor_credentials.get("access_token", "")
        if visitor_api_key == "":
            raise ValueError("Unable to retrieve token.")

        return Client(url=self.cfg.url, api_key=visitor_api_key)

    @property
    def content(self) -> Content:
        """
        The content resource interface.

        Returns
        -------
        Content
            The content resource instance.
        """
        return Content(self._ctx)

    @property
    @requires(version="2023.05.0")
    def environments(self) -> Environments:
        return _ResourceSequence(self._ctx, "v1/environments")

    @property
    def groups(self) -> Groups:
        """The groups resource interface.

        Returns
        -------
        Groups
            The groups resource interface.
        """
        return Groups(self._ctx)

    @property
    def me(self) -> User:
        """
        The connected user.

        Returns
        -------
        User
            The currently authenticated user.
        """
        return me.get(self._ctx)

    @property
    def metrics(self) -> Metrics:
        """
        The Metrics API interface.

        The Metrics API is designed for capturing, retrieving, and managing
        quantitative measurements of Connect interactions. It is commonly used
        for monitoring and analyzing system performance, user behavior, and
        business processes. This API facilitates real-time data collection and
        accessibility, enabling organizations to make informed decisions based
        on key performance indicators (KPIs).

        Returns
        -------
        Metrics
            The metrics API instance.

        Examples
        --------
        >>> from posit import connect
        >>> client = connect.Client()
        >>> content_guid = "2243770d-ace0-4782-87f9-fe2aeca14fc8"
        >>> events = client.metrics.usage.find(content_guid=content_guid)
        >>> len(events)
        24
        """
        return Metrics(self._ctx)

    @property
    @requires(version="2024.08.0")
    def oauth(self) -> OAuth:
        """
        The OAuth API interface.

        Returns
        -------
        OAuth
            The oauth API instance.
        """
        return OAuth(self._ctx, self.cfg.api_key)

    @property
    @requires(version="2024.11.0")
    def packages(self) -> Packages:
        return _PaginatedResourceSequence(
            self._ctx, "v1/packages", uid="name", page_size=1_000_000
        )

    @property
    def system(self) -> System:
        return System(self._ctx, "v1/system")

    @property
    def tags(self) -> Tags:
        """
        The tags resource interface.

        Returns
        -------
        Tags
            The tags resource instance.

        Examples
        --------
        ```python
        import posit

        client = posit.connect.Client()

        tags = client.tags.find()
        ```
        """
        return Tags(self._ctx, "v1/tags")

    @property
    def tasks(self) -> Tasks:
        """
        The tasks resource interface.

        Returns
        -------
        tasks.Tasks
            The tasks resource instance.
        """
        return Tasks(self._ctx)

    @property
    def users(self) -> Users:
        """
        The users resource interface.

        Returns
        -------
        Users
            The users resource instance.
        """
        return Users(self._ctx)

    @property
    def vanities(self) -> Vanities:
        return Vanities(self._ctx)

    @property
    def version(self) -> str | None:
        """
        The server version.

        Returns
        -------
        str
            The version of the Posit Connect server.
        """
        return self._ctx.version

    def __del__(self):
        """Close the session when the Client instance is deleted."""
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def __enter__(self):
        """Enter method for using the client as a context manager."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Close the session if it exists.

        Parameters
        ----------
        exc_type : type
            The type of the exception raised (if any).
        exc_value : Exception
            The exception instance raised (if any).
        exc_tb : traceback
            The traceback for the exception raised (if any).
        """
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def request(self, method: str, path: str, **kwargs) -> Response:
        """
        Send an HTTP request.

        A facade for [](`requests.request`) configured for the target server.

        Parameters
        ----------
        method : str
            The HTTP method to use for the request.
        path : str
            Appended to the url object attribute.
        **kwargs
            Additional keyword arguments passed to [](`requests.request`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.request(method, url, **kwargs)

    def get(self, path: str, **kwargs) -> Response:
        """
        Send a GET request.

        A facade for [](`requests.get`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.get`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.get(url, **kwargs)

    def post(self, path: str, **kwargs) -> Response:
        """
        Send a POST request.

        A facade for [](`requests.post`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.post`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.post(url, **kwargs)

    def put(self, path: str, **kwargs) -> Response:
        """
        Send a PUT request.

        A facade for [](`requests.put`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.put`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.put(url, **kwargs)

    def patch(self, path: str, **kwargs) -> Response:
        """
        Send a PATCH request.

        A facade for [](`requests.patch`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.patch`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.patch(url, **kwargs)

    def delete(self, path: str, **kwargs) -> Response:
        """
        Send a DELETE request.

        A facade for [](`requests.delete`) configured for the target server.

        Parameters
        ----------
        path : str
            Appended to the configured base url.
        **kwargs
            Additional keyword arguments passed to [](`requests.delete`).

        Returns
        -------
        Response
            A [](`requests.Response`) object.
        """
        url = self.cfg.url + path
        return self.session.delete(url, **kwargs)



from pathlib import Path

import pytest
from packaging import version

from posit.connect import Client
from posit.connect.system import SystemRuntimeCache
from posit.connect.tasks import Task

from . import CONNECT_VERSION


@pytest.mark.skipif(
    CONNECT_VERSION < version.parse("2024.05.0"),
    reason="Cache runtimes not implemented",
)
class TestSystem:
    @classmethod
    def setup_class(cls):
        cls.client = Client()
        cls.content = cls.client.content.create(name=cls.__name__)
        path = Path("../../../resources/connect/bundles/example-flask-minimal/bundle.tar.gz")
        path = (Path(__file__).parent / path).resolve()
        bundle = cls.content.bundles.create(str(path))
        task = bundle.deploy()
        task.wait_for()

    @classmethod
    def teardown_class(cls):
        cls.content.delete()

    def test_runtime_caches(self):
        runtimes = self.client.system.caches.runtime.find()
        assert len(runtimes) > 0

    def test_runtime_cache_destroy(self):
        # Find existing runtime caches
        runtimes: list[SystemRuntimeCache] = self.client.system.caches.runtime.find()
        assert len(runtimes) > 0

        # Get the first cache for testing
        cache = runtimes[0]
        assert isinstance(cache, SystemRuntimeCache)

        # Test dry run destroy (should return None and not actually destroy)
        result = cache.destroy(dry_run=True)
        assert result is None

        # Verify cache still exists after dry run
        runtimes_after_dry_run = self.client.system.caches.runtime.find()
        assert len(runtimes_after_dry_run) == len(runtimes)

        # Test actual destroy
        task: Task = cache.destroy()
        assert isinstance(task, Task)
        task.wait_for()

        # Verify cache was removed
        runtimes_after_destroy = self.client.system.caches.runtime.find()
        assert len(runtimes_after_destroy) == len(runtimes) - 1



"""OAuth association resources."""

from __future__ import annotations

from functools import partial

from typing_extensions import TYPE_CHECKING, List, Optional, overload

from ..resources import BaseResource, Resources, _matches_exact, _matches_pattern

if TYPE_CHECKING:
    from ..context import Context
    from ..oauth import types


class Association(BaseResource):
    pass


class IntegrationAssociations(Resources):
    """IntegrationAssociations resource."""

    def __init__(self, ctx: Context, integration_guid: str) -> None:
        super().__init__(ctx)
        self.integration_guid = integration_guid

    def find(self) -> List[Association]:
        """Find OAuth associations.

        Returns
        -------
        List[Association]
        """
        path = f"v1/oauth/integrations/{self.integration_guid}/associations"
        response = self._ctx.client.get(path)
        return [
            Association(
                self._ctx,
                **result,
            )
            for result in response.json()
        ]


class ContentItemAssociations(Resources):
    """ContentItemAssociations resource."""

    def __init__(self, ctx, content_guid: str):
        super().__init__(ctx)
        self.content_guid = content_guid

    def find(self) -> List[Association]:
        """Find OAuth associations.

        Returns
        -------
        List[Association]
        """
        path = f"v1/content/{self.content_guid}/oauth/integrations/associations"
        response = self._ctx.client.get(path)
        return [
            Association(
                self._ctx,
                **result,
            )
            for result in response.json()
        ]

    def find_by(
        self,
        integration_type: Optional[types.OAuthIntegrationType | str] = None,
        auth_type: Optional[types.OAuthIntegrationAuthType | str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        guid: Optional[str] = None,
    ) -> Association | None:
        """Find an OAuth integration associated with content by various criteria.

        Parameters
        ----------
        integration_type : Optional[types.OAuthIntegrationType | str]
            The type of the integration (e.g., "aws", "azure").
        auth_type : Optional[types.OAuthIntegrationAuthType | str]
            The authentication type of the integration (e.g., "Viewer", "Service Account").
        name : Optional[str]
            A regex pattern to match the integration name. For exact matches, use `^` and `$`. For example,
            `^My Integration$` will match only "My Integration".
        description : Optional[str]
            A regex pattern to match the integration description. For exact matches, use `^` and `$`. For example,
            `^My Integration Description$` will match only "My Integration Description".
        guid : Optional[str]
            The unique identifier of the integration.

        Returns
        -------
        Association | None
            The first matching association, or None if no match is found.
        """
        filters = []
        if integration_type is not None:
            filters.append(
                partial(_matches_exact, key="oauth_integration_template", value=integration_type)
            )
        if auth_type is not None:
            filters.append(
                partial(_matches_exact, key="oauth_integration_auth_type", value=auth_type)
            )
        if name is not None:
            filters.append(partial(_matches_pattern, key="oauth_integration_name", pattern=name))
        if description is not None:
            filters.append(
                partial(_matches_pattern, key="oauth_integration_description", pattern=description)
            )
        if guid is not None:
            filters.append(partial(_matches_exact, key="oauth_integration_guid", value=guid))

        for association in self.find():
            if all(f(association) for f in filters):
                return association

        return None

    def delete(self) -> None:
        """Delete integration associations."""
        data = []

        path = f"v1/content/{self.content_guid}/oauth/integrations/associations"
        self._ctx.client.put(path, json=data)

    @overload
    def update(self, integration_guid: str) -> None:
        """Set a single integration association.

        Parameters
        ----------
        integration_guid : str
            The unique identifier of the integration.
        """

    @overload
    def update(self, integration_guid: list[str]) -> None:
        """Set multiple integration associations.

        Parameters
        ----------
        integration_guids : list[str]
            A list of unique identifiers of the integrations.
        """

    def update(self, integration_guid: str | list[str]) -> None:
        """Set integration associations."""
        if isinstance(integration_guid, str):
            integration_guids = [integration_guid]
        else:
            integration_guids = integration_guid

        data = [{"oauth_integration_guid": guid} for guid in integration_guids]

        path = f"v1/content/{self.content_guid}/oauth/integrations/associations"
        self._ctx.client.put(path, json=data)



from __future__ import annotations

import pytest

from posit.workbench.external.databricks import (
    POSIT_WORKBENCH_AUTH_TYPE,
    WorkbenchStrategy,
)

try:
    from databricks.sdk.core import Config
except ImportError:
    pytestmark = pytest.mark.skipif(True, reason="requires the Databricks SDK")


class TestPositCredentialsHelpers:
    def test_default_workbench_strategy(self):
        # By default, the WorkbenchStrategy should use "workbench" profile.
        strategy = WorkbenchStrategy()

        with pytest.raises(ValueError, match="profile=workbench"):
            strategy()

    def test_workbench_strategy(self):
        # providing a Config is allowed
        cs = WorkbenchStrategy(
            config=Config(host="https://databricks.com/workspace", token="token")  # pyright: ignore[reportPossiblyUnboundVariable]
        )
        assert cs.auth_type() == POSIT_WORKBENCH_AUTH_TYPE
        cp = cs()

        # token from the Config is passed through to the auth header
        assert cp() == {"Authorization": "Bearer token"}



from enum import Enum

GRANT_TYPE = "urn:ietf:params:oauth:grant-type:token-exchange"


class OAuthTokenType(str, Enum):
    ACCESS_TOKEN = "urn:ietf:params:oauth:token-type:access_token"
    AWS_CREDENTIALS = "urn:ietf:params:aws:token-type:credentials"
    API_KEY = "urn:posit:connect:api-key"
    CONTENT_SESSION_TOKEN = "urn:posit:connect:content-session-token"
    USER_SESSION_TOKEN = "urn:posit:connect:user-session-token"


class OAuthIntegrationAuthType(str, Enum):
    """OAuth integration authentication type."""

    VIEWER = "Viewer"
    SERVICE_ACCOUNT = "Service Account"
    VISITOR_API_KEY = "Visitor API Key"


class OAuthIntegrationType(str, Enum):
    """OAuth integration type."""

    AWS = "aws"
    AZURE = "azure"
    AZURE_OPENAI = "azure-openai"
    CONNECT = "connect"
    CUSTOM = "custom"
    DATABRICKS = "databricks"
    GITHUB = "github"
    GOOGLE_BIGQUERY = "bigquery"
    GOOGLE_DRIVE = "drive"
    GOOGLE_SHEETS = "sheets"
    GOOGLE_VERTEX_AI = "vertex-ai"
    MSGRAPH = "msgraph"
    SALESFORCE = "salesforce"
    SHAREPOINT = "sharepoint"
    SNOWFLAKE = "snowflake"



import time
from pathlib import Path

import pytest
from packaging import version

from posit import connect

from . import CONNECT_VERSION


class TestBundles:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create(
            name=f"test-bundles-{int(time.time())}",
            title="Test Bundles",
            access_type="all",
        )
        # Path to the test bundle
        bundle_path = Path(
            "../../../resources/connect/bundles/example-flask-minimal/bundle.tar.gz"
        )
        cls.bundle_path = (Path(__file__).parent / bundle_path).resolve()

    @classmethod
    def teardown_class(cls):
        cls.content.delete()

    def test_create_bundle(self):
        """Test creating a bundle."""
        bundle = self.content.bundles.create(str(self.bundle_path))
        assert bundle["id"]
        assert bundle["content_guid"] == self.content["guid"]

    def test_find_bundles(self):
        """Test finding all bundles."""
        # Create a bundle first
        self.content.bundles.create(str(self.bundle_path))

        # Find all bundles
        bundles = self.content.bundles.find()
        assert len(bundles) >= 1

    def test_find_one_bundle(self):
        """Test finding a single bundle."""
        # Create a bundle first
        self.content.bundles.create(str(self.bundle_path))

        # Find one bundle
        bundle = self.content.bundles.find_one()
        assert bundle is not None
        assert bundle["content_guid"] == self.content["guid"]

    def test_get_bundle(self):
        """Test getting a specific bundle."""
        # Create a bundle first
        created_bundle = self.content.bundles.create(str(self.bundle_path))

        # Get the bundle by ID
        bundle = self.content.bundles.get(created_bundle["id"])
        assert bundle["id"] == created_bundle["id"]

    @pytest.mark.skipif(
        CONNECT_VERSION < version.parse("2025.02.0"), reason="Requires Connect 2025.02.0 or later"
    )
    def test_active_bundle(self):
        """Test retrieving the active bundle."""
        # Initially, no bundle should be active
        assert self.content.bundles.active() is None

        # Create and deploy a bundle
        bundle = self.content.bundles.create(str(self.bundle_path))
        task = bundle.deploy()
        task.wait_for()

        # Wait for the bundle to become active
        max_retries = 10
        active_bundle = None
        for _ in range(max_retries):
            active_bundle = self.content.bundles.active()
            if active_bundle is not None:
                break
            time.sleep(1)

        # Verify the bundle is now active
        assert active_bundle is not None
        assert active_bundle["id"] == bundle["id"]
        assert active_bundle.get("active") is True

        # Create another bundle but don't deploy it
        bundle2 = self.content.bundles.create(str(self.bundle_path))

        # Verify the active bundle is still the first one
        active_bundle = self.content.bundles.active()
        assert active_bundle is not None
        assert active_bundle["id"] == bundle["id"]
        assert active_bundle["id"] != bundle2["id"]



from __future__ import annotations

import os

from typing_extensions import TYPE_CHECKING, Optional, TypedDict

from ..oauth import types
from ..resources import Resources
from .integrations import Integrations
from .sessions import Sessions

if TYPE_CHECKING:
    from ..context import Context

# Adding constants for backwards compatibility
# Moving these could break existing code that imports them directly
GRANT_TYPE = types.GRANT_TYPE

OAuthTokenType = types.OAuthTokenType


def _get_content_session_token() -> str:
    """Return the content session token.

    Reads the environment variable 'CONNECT_CONTENT_SESSION_TOKEN'.

    Raises
    ------
        ValueError: If CONNECT_CONTENT_SESSION_TOKEN is not set or invalid

    Returns
    -------
        str
    """
    value = os.environ.get("CONNECT_CONTENT_SESSION_TOKEN")
    if not value:
        raise ValueError(
            "Invalid value for 'CONNECT_CONTENT_SESSION_TOKEN': Must be a non-empty string."
        )
    return value


class OAuth(Resources):
    def __init__(self, ctx: Context, api_key: str) -> None:
        super().__init__(ctx)
        self.api_key = api_key
        self._path = "v1/oauth/integrations/credentials"

    @property
    def integrations(self):
        return Integrations(self._ctx)

    @property
    def sessions(self):
        return Sessions(self._ctx)

    def get_credentials(
        self,
        user_session_token: str,
        requested_token_type: Optional[str | types.OAuthTokenType] = None,
        audience: Optional[str] = None,
    ) -> Credentials:
        """Perform an oauth credential exchange with a user-session-token.

        Parameters
        ----------
        user_session_token : str
            The user session token to use for the exchange.
        requested_token_type : str or OAuthTokenType, optional
            The type of token being requested. This can be one of the predefined types in `OAuthTokenType` or a custom string.
        audience : str, optional
            The intended audience for the token. This must be a valid integration GUID.

        Returns
        -------
        Credentials
            The credentials obtained from the exchange.
        """
        # craft a credential exchange request
        data = {}
        data["grant_type"] = types.GRANT_TYPE
        data["subject_token_type"] = types.OAuthTokenType.USER_SESSION_TOKEN
        data["subject_token"] = user_session_token
        if requested_token_type:
            data["requested_token_type"] = requested_token_type
        if audience:
            data["audience"] = audience

        response = self._ctx.client.post(self._path, data=data)
        return Credentials(**response.json())

    def get_content_credentials(
        self,
        content_session_token: Optional[str] = None,
        requested_token_type: Optional[str | types.OAuthTokenType] = None,
        audience: Optional[str] = None,
    ) -> Credentials:
        """Perform an oauth credential exchange with a content-session-token.

        Parameters
        ----------
        content_session_token : str, optional
            The content session token to use for the exchange. If not provided, the function will attempt to read the token from the environment variable 'CONNECT_CONTENT_SESSION_TOKEN'.
        requested_token_type : str or OAuthTokenType, optional
            The type of token being requested. This can be one of the predefined types in `OAuthTokenType` or a custom string.
        audience : str, optional
            The intended audience for the token. This must be a valid integration GUID.

        Returns
        -------
        Credentials
            The credentials obtained from the exchange.

        """
        # craft a credential exchange request
        data = {}
        data["grant_type"] = types.GRANT_TYPE
        data["subject_token_type"] = types.OAuthTokenType.CONTENT_SESSION_TOKEN
        data["subject_token"] = content_session_token or _get_content_session_token()
        if requested_token_type:
            data["requested_token_type"] = requested_token_type
        if audience:
            data["audience"] = audience

        response = self._ctx.client.post(self._path, data=data)
        return Credentials(**response.json())


class Credentials(TypedDict, total=False):
    access_token: str
    issued_token_type: str
    token_type: str



from __future__ import annotations

from typing_extensions import (
    Iterable,
    Protocol,
)

from ..resources import Resource, ResourceSequence, _ResourceSequence
from .rename_params import rename_params


class Hit(Resource, Protocol):
    pass


class Hits(ResourceSequence[Hit], Protocol):
    def fetch(
        self,
        *,
        start: str = ...,
        end: str = ...,
    ) -> Iterable[Hit]:
        """
        Fetch all content hit records matching the specified conditions.

        Parameters
        ----------
        start : str, not required
            The timestamp that starts the time window of interest in RFC 3339 format.
            Any hit information that happened prior to this timestamp will not be returned.
            Example: "2025-05-01T00:00:00Z"
        end : str, not required
            The timestamp that ends the time window of interest in RFC 3339 format.
            Any hit information that happened after this timestamp will not be returned.
            Example: "2025-05-02T00:00:00Z"

        Returns
        -------
        Iterable[Hit]
            All content hit records matching the specified conditions.
        """
        ...

    def find_by(
        self,
        *,
        id: str = ...,  # noqa: A002
        content_guid: str = ...,
        user_guid: str = ...,
        timestamp: str = ...,
    ) -> Hit | None:
        """
        Find the first hit record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        id : str, not required
            The ID of the activity record.
        content_guid : str, not required
            The GUID, in RFC4122 format, of the content this information pertains to.
        user_guid : str, not required
            The GUID, in RFC4122 format, of the user that visited the content.
            May be null when the target content does not require a user session.
        timestamp : str, not required
            The timestamp, in RFC 3339 format, when the user visited the content.

        Returns
        -------
        Hit | None
            The first hit record matching the specified conditions, or `None` if no such record exists.
        """
        ...


class _Hits(_ResourceSequence, Hits):
    def fetch(
        self,
        **kwargs,
    ) -> Iterable[Hit]:
        """
        Fetch all content hit records matching the specified conditions.

        Parameters
        ----------
        start : str, not required
            The timestamp that starts the time window of interest in RFC 3339 format.
            Any hit information that happened prior to this timestamp will not be returned.
            This corresponds to the `from` parameter in the API.
            Example: "2025-05-01T00:00:00Z"
        end : str, not required
            The timestamp that ends the time window of interest in RFC 3339 format.
            Any hit information that happened after this timestamp will not be returned.
            This corresponds to the `to` parameter in the API.
            Example: "2025-05-02T00:00:00Z"

        Returns
        -------
        Iterable[Hit]
            All content hit records matching the specified conditions.
        """
        params = rename_params(kwargs)
        return super().fetch(**params)



from urllib.parse import urljoin

import requests


class Session(requests.Session):
    """Custom session that implements CURLOPT_POSTREDIR.

    This class mimics the functionality of CURLOPT_POSTREDIR from libcurl by
    providing a custom implementation of the POST method. It allows the caller
    to control whether the original POST data is preserved on redirects or if the
    request should be converted to a GET when a redirect occurs. This is achieved
    by disabling automatic redirect handling and manually following the redirect
    chain with the desired behavior.

    Notes
    -----
    The custom `post` method in this class:

    - Disables automatic redirect handling by setting ``allow_redirects=False``.
    - Manually follows redirects up to a specified ``max_redirects``.
    - Determines the HTTP method for subsequent requests based on the response
      status code and the ``preserve_post`` flag:

        - For HTTP status codes 307 and 308, the method and request body are
          always preserved as POST.
        - For other redirects (e.g., 301, 302, 303), the behavior is determined
          by ``preserve_post``:
            - If ``preserve_post=True``, the POST method is maintained.
            - If ``preserve_post=False``, the method is converted to GET and the
              request body is discarded.

    Examples
    --------
    Create a session and send a POST request while preserving POST data on redirects:

    >>> session = Session()
    >>> response = session.post(
    ...     "https://example.com/api", data={"key": "value"}, preserve_post=True
    ... )
    >>> print(response.status_code)

    See Also
    --------
    requests.Session : The base session class from the requests library.
    """

    def post(self, url, data=None, json=None, preserve_post=True, max_redirects=5, **kwargs):
        """
        Send a POST request and handle redirects manually.

        Parameters
        ----------
        url : str
            The URL to send the POST request to.
        data : dict, bytes, or file-like object, optional
            The form data to send.
        json : any, optional
            The JSON data to send.
        preserve_post : bool, optional
            If True, re-send POST data on redirects (mimicking CURLOPT_POSTREDIR);
            if False, converts to GET on 301/302/303 responses.
        max_redirects : int, optional
            Maximum number of redirects to follow.
        **kwargs
            Additional keyword arguments passed to the request.

        Returns
        -------
        requests.Response
            The final response after following redirects.
        """
        # Force manual redirect handling by disabling auto redirects.
        kwargs["allow_redirects"] = False

        # Initial POST request
        response = super().post(url, data=data, json=json, **kwargs)
        redirect_count = 0

        # Manually follow redirects, if any
        while response.is_redirect and redirect_count < max_redirects:
            redirect_url = response.headers.get("location")
            if not redirect_url:
                break  # No redirect URL; exit loop

            redirect_url = urljoin(response.url, redirect_url)

            # For 307 and 308 the HTTP spec mandates preserving the method and body.
            if response.status_code in (307, 308):
                method = "POST"
            else:
                if preserve_post:
                    method = "POST"
                else:
                    method = "GET"
                    data = None
                    json = None

            # Perform the next request in the redirect chain.
            response = self.request(method, redirect_url, data=data, json=json, **kwargs)
            redirect_count += 1

        return response



import pytest
import responses

from posit.connect.sessions import Session


@responses.activate
def test_post_no_redirect():
    url = "https://connect.example.com/api"
    responses.add(responses.POST, url, json={"result": "ok"}, status=200)

    session = Session()
    response = session.post(url, data={"key": "value"})

    assert response.status_code == 200
    assert len(responses.calls) == 1
    # Confirm that the request method was POST.
    assert responses.calls[0].request.method == "POST"


@responses.activate
def test_post_with_redirect_preserve():
    initial_url = "http://connect.example.com/api"
    redirect_url = "http://connect.example.com/redirect"

    responses.add(responses.POST, initial_url, status=302, headers={"location": "/redirect"})
    responses.add(responses.POST, redirect_url, json={"result": "redirected"}, status=200)

    session = Session()
    response = session.post(initial_url, data={"key": "value"}, preserve_post=True)

    assert response.status_code == 200
    assert len(responses.calls) == 2

    # Both calls should use the POST method.
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[1].request.method == "POST"


@responses.activate
def test_post_with_redirect_no_preserve():
    initial_url = "http://connect.example.com/api"
    redirect_url = "http://connect.example.com/redirect"

    responses.add(responses.POST, initial_url, status=302, headers={"location": "/redirect"})
    responses.add(responses.GET, redirect_url, json={"result": "redirected"}, status=200)

    session = Session()
    response = session.post(initial_url, data={"key": "value"}, preserve_post=False)

    assert response.status_code == 200
    assert len(responses.calls) == 2
    # The initial call is a POST, but the follow-up should be a GET since preserve_post is False
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[1].request.method == "GET"


@pytest.mark.parametrize("status_code", [307, 308])
@responses.activate
def test_post_redirect_307_308(status_code):
    initial_url = "http://connect.example.com/api"
    redirect_url = "http://connect.example.com/redirect"

    # For 307 and 308 redirects, the HTTP spec mandates preserving the method.
    responses.add(
        responses.POST, initial_url, status=status_code, headers={"location": "/redirect"}
    )
    responses.add(responses.POST, redirect_url, json={"result": "redirected"}, status=200)

    session = Session()
    # Even with preserve_post=False, a 307 or 308 redirect should use POST.
    response = session.post(initial_url, data={"key": "value"}, preserve_post=False)

    assert response.status_code == 200
    assert len(responses.calls) == 2
    # Confirm that the method for the redirect is still POST.
    assert responses.calls[1].request.method == "POST"


@responses.activate
def test_post_redirect_max_redirects():
    initial_url = "http://connect.example.com/api"
    redirect1_url = "http://connect.example.com/redirect1"
    redirect2_url = "http://connect.example.com/redirect2"

    # Build a chain of 3 redirects.
    responses.add(responses.POST, initial_url, status=302, headers={"location": "/redirect1"})
    responses.add(responses.POST, redirect1_url, status=302, headers={"location": "/redirect2"})
    responses.add(responses.POST, redirect2_url, status=302, headers={"location": "/redirect3"})

    session = Session()
    # Limit to 2 redirects; thus, the third redirect response should not be followed.
    response = session.post(
        initial_url, data={"key": "value"}, max_redirects=2, preserve_post=True
    )

    # The calls should include: initial, first redirect, and second redirect.
    assert len(responses.calls) == 3
    # The final response is the one from the second redirect.
    assert response.status_code == 302
    # The Location header should point to the third URL.
    assert response.headers.get("location") == "/redirect3"


@responses.activate
def test_post_redirect_no_location():
    url = "http://connect.example.com/api"
    # Simulate a redirect response that lacks a Location header.
    responses.add(responses.POST, url, status=302, headers={})

    session = Session()
    response = session.post(url, data={"key": "value"})

    # The loop should break immediately since there is no location to follow.
    assert len(responses.calls) == 1
    assert response.status_code == 302


@responses.activate
def test_post_redirect_location_none_explicit():
    url = "http://connect.example.com/api"

    # Use a callback to explicitly return a None for the "location" header.
    def request_callback(request):
        return (302, {"location": ""}, "Redirect without location")

    responses.add_callback(responses.POST, url, callback=request_callback)

    session = Session()
    response = session.post(url, data={"key": "value"})

    # The redirect loop should break since location is None.
    assert len(responses.calls) == 1
    assert response.status_code == 302



from __future__ import annotations

import base64
import json
from datetime import datetime

from typing_extensions import TYPE_CHECKING, Optional, TypedDict

from ..oauth.types import OAuthTokenType

if TYPE_CHECKING:
    from ..client import Client


class Credentials(TypedDict):
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str
    expiration: datetime


def get_credentials(
    client: Client,
    user_session_token: str,
    audience: Optional[str] = None,
) -> Credentials:
    """
    Get AWS credentials using OAuth token exchange for an AWS Viewer integration.

    According to RFC 8693, the access token must be a base64-encoded JSON object
    containing the AWS credentials. This function will return the decoded and
    deserialized AWS credentials.

    Examples
    --------
    ```python
    from posit.connect import Client
    from posit.connect.external.aws import get_aws_credentials
    import boto3
    from shiny.express import session

    client = Client()
    session_token = session.http_conn.headers.get("Posit-Connect-User-Session-Token")
    credentials = get_aws_credentials(client, user_session_token)
    aws_session_expiration = credentials["expiration"]
    aws_session = boto3.Session(
        aws_access_key_id=credentials["aws_access_key_id"],
        aws_secret_access_key=credentials["aws_secret_access_key"],
        aws_session_token=credentials["aws_session_token"],
    )

    s3 = aws_session.resource("s3")
    bucket = s3.Bucket("your-bucket-name")
    ```

    Parameters
    ----------
    client : Client
        The client to use for making requests
    user_session_token : str
        The user session token to exchange

    Returns
    -------
    Dict[str, str]
        Dictionary containing AWS credentials with keys:
        access_key_id, secret_access_key, session_token, and expiration
    """
    # Get credentials using OAuth
    credentials = client.oauth.get_credentials(
        user_session_token=user_session_token,
        requested_token_type=OAuthTokenType.AWS_CREDENTIALS,
        audience=audience,
    )

    # Decode base64 access token
    access_token = credentials.get("access_token")
    if not access_token:
        raise ValueError("No access token found in credentials")
    return _decode_access_token(access_token)


def get_content_credentials(
    client: Client,
    content_session_token: Optional[str] = None,
    audience: Optional[str] = None,
) -> Credentials:
    """
    Get AWS credentials using OAuth token exchange for an AWS Service Account integration.

    According to RFC 8693, the access token must be a base64-encoded JSON object
    containing the AWS credentials. This function will return the decoded and
    deserialized AWS credentials.

    Examples
    --------
    ```python
    from posit.connect import Client
    from posit.connect.external.aws import get_aws_content_credentials
    import boto3

    client = Client()
    credentials = get_aws_content_credentials(client)
    session_expiration = credentials["expiration"]
    aws_session = boto3.Session(
        aws_access_key_id=credentials["aws_access_key_id"],
        aws_secret_access_key=credentials["aws_secret_access_key"],
        aws_session_token=credentials["aws_session_token"],
    )

    s3 = session.resource("s3")
    bucket = s3.Bucket("your-bucket-name")
    ```

    Parameters
    ----------
    client : Client
        The client to use for making requests
    content_session_token : str
        The content session token to exchange

    Returns
    -------
    Dict[str, str]
        Dictionary containing AWS credentials with keys:
        access_key_id, secret_access_key, session_token, and expiration
    """
    # Get credentials using OAuth
    credentials = client.oauth.get_content_credentials(
        content_session_token=content_session_token,
        requested_token_type=OAuthTokenType.AWS_CREDENTIALS,
        audience=audience,
    )

    # Decode base64 access token
    access_token = credentials.get("access_token")
    if not access_token:
        raise ValueError("No access token found in credentials")
    return _decode_access_token(access_token)


def _decode_access_token(access_token: str) -> Credentials:
    """
    Decode and deserialize an access token containing AWS credentials.

    According to RFC 8693, the access token must be a base64-encoded JSON object
    containing the AWS credentials. This function will decode and deserialize the
    access token and return the AWS credentials.

    Parameters
    ----------
    access_token : str
        The access token to decode

    Returns
    -------
    Credentials
        Dictionary containing AWS credentials with keys:
        access_key_id, secret_access_key, session_token, and expiration
    """
    decoded_bytes = base64.b64decode(access_token)
    decoded_str = decoded_bytes.decode("utf-8")
    aws_credentials = json.loads(decoded_str)

    return Credentials(
        aws_access_key_id=aws_credentials["accessKeyId"],
        aws_secret_access_key=aws_credentials["secretAccessKey"],
        aws_session_token=aws_credentials["sessionToken"],
        expiration=datetime.strptime(aws_credentials["expiration"], "%Y-%m-%dT%H:%M:%SZ"),
    )



from __future__ import annotations

from unittest.mock import patch

import pytest
import responses
from typing_extensions import Dict

from posit.connect import Client
from posit.connect.external.databricks import (
    POSIT_OAUTH_INTEGRATION_AUTH_TYPE,
    ConnectStrategy,
    _new_bearer_authorization_header,
    _PositConnectContentCredentialsProvider,
    _PositConnectViewerCredentialsProvider,
    databricks_config,
)
from posit.connect.oauth import Credentials

try:
    from databricks.sdk.core import Config, DefaultCredentials
    from databricks.sdk.credentials_provider import (
        CredentialsProvider,
        CredentialsStrategy,
    )

    # construct a DefaultCredentials CredentialsStrategy
    # weirdly, you have to call `__call__()` at least once in order to initialize `auth_type()`
    # This is the expected credentials strategy when none is provided to our databricks_config() helper
    expected_credentials = DefaultCredentials()  # pyright: ignore[reportPossiblyUnboundVariable]
    expected_credentials(Config(auth_type="pat", token="asdf", host="https://databricks.com/"))  # pyright: ignore[reportPossiblyUnboundVariable]

except ImportError:
    pytestmark = pytest.mark.skipif(True, reason="requires the Databricks SDK")


class mock_strategy(CredentialsStrategy):  # pyright: ignore[reportPossiblyUnboundVariable]
    def __init__(self, name: str):
        self.name = name

    def auth_type(self) -> str:
        return self.name

    def __call__(self, *args, **kwargs) -> CredentialsProvider:
        def inner() -> Dict[str, str]:
            return {"Authorization": f"Bearer {self.name}"}

        return inner


def register_mocks():
    responses.post(
        "https://connect.example/__api__/v1/oauth/integrations/credentials",
        match=[
            responses.matchers.urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                    "subject_token_type": "urn:posit:connect:user-session-token",
                    "subject_token": "cit",
                },
            ),
        ],
        json={
            "access_token": "dynamic-viewer-access-token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "Bearer",
        },
    )

    responses.post(
        "https://connect.example/__api__/v1/oauth/integrations/credentials",
        match=[
            responses.matchers.urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                    "subject_token_type": "urn:posit:connect:content-session-token",
                    "subject_token": "cit",
                },
            ),
        ],
        json={
            "access_token": "content-access-token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "Bearer",
        },
    )


class TestPositCredentialsHelpers:
    def test_new_bearer_authorization_header(self):
        credential = Credentials()
        credential["token_type"] = "token_type"
        credential["issued_token_type"] = "issued_token_type"

        with pytest.raises(ValueError):
            _new_bearer_authorization_header(credential)

        credential["access_token"] = "access_token"
        result = _new_bearer_authorization_header(credential)
        assert result == {"Authorization": "Bearer access_token"}

    @patch.dict("os.environ", {"CONNECT_CONTENT_SESSION_TOKEN": "cit"})
    @responses.activate
    def test_posit_content_credentials_provider(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cp = _PositConnectContentCredentialsProvider(client=client)
        assert cp() == {"Authorization": "Bearer content-access-token"}

    @responses.activate
    def test_posit_credentials_provider(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cp = _PositConnectViewerCredentialsProvider(client=client, user_session_token="cit")
        assert cp() == {"Authorization": "Bearer dynamic-viewer-access-token"}

    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    @patch.dict("os.environ", {"CONNECT_CONTENT_SESSION_TOKEN": "cit"})
    @responses.activate
    def test_connect_strategy(self):
        register_mocks()
        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None

        # the default implementation uses Service Account authentication
        cs = ConnectStrategy(client=client)
        assert cs.auth_type() == POSIT_OAUTH_INTEGRATION_AUTH_TYPE
        cp = cs()
        assert cp() == {"Authorization": "Bearer content-access-token"}

        # if a session token is provided then Viewer auth is used
        cs = ConnectStrategy(client=client, user_session_token="cit")
        cp = cs()
        assert cp() == {"Authorization": "Bearer dynamic-viewer-access-token"}

    def test_databricks_config(self):
        # credentials_strategy is removed if it is provided
        cfg = databricks_config(
            credentials_strategy=mock_strategy("mock"),
            auth_type="pat",
            token="asdf",
            host="https://databricks.com/",
        )
        assert cfg._credentials_strategy.auth_type() == expected_credentials.auth_type()

        # kwargs are passed through to the Config() constructor
        cfg = databricks_config(
            host="https://databricks.com",
            cluster_id="cluster_id",
            warehouse_id="warehouse_id",
            token="token",
        )
        assert cfg.host == "https://databricks.com"
        assert cfg.cluster_id == "cluster_id"
        assert cfg.warehouse_id == "warehouse_id"
        assert cfg.token == "token"

    def test_databricks_config_default(self):
        cfg = databricks_config(
            posit_default_strategy=mock_strategy("default"),
            posit_workbench_strategy=mock_strategy("workbench"),
            posit_connect_strategy=mock_strategy("connect"),
        )
        assert cfg._credentials_strategy.auth_type() == "default"

        # default fallback defaults to DefaultCredentials() when none is provided
        cfg = databricks_config(auth_type="pat", token="asdf", host="https://databricks.com/")
        assert cfg._credentials_strategy.auth_type() == expected_credentials.auth_type()

    @patch.dict("os.environ", {"RS_SERVER_ADDRESS": "https://workbench.posit.co/"})
    def test_databricks_config_workbench(self):
        cfg = databricks_config(
            posit_default_strategy=mock_strategy("default"),
            posit_workbench_strategy=mock_strategy("workbench"),
            posit_connect_strategy=mock_strategy("connect"),
        )
        assert cfg._credentials_strategy.auth_type() == "workbench"

        # workbench defaults to DefaultCredentials() when none is provided
        cfg = databricks_config(auth_type="pat", token="asdf", host="https://databricks.com/")
        assert cfg._credentials_strategy.auth_type() == expected_credentials.auth_type()

    @patch.dict("os.environ", {"CONNECT_API_KEY": "API_KEY"})
    @patch.dict("os.environ", {"CONNECT_SERVER": "https://connect.posit.co/"})
    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    def test_databricks_config_connect(self):
        cfg = databricks_config(
            posit_default_strategy=mock_strategy("default"),
            posit_workbench_strategy=mock_strategy("workbench"),
            posit_connect_strategy=mock_strategy("connect"),
        )
        assert cfg._credentials_strategy.auth_type() == "connect"

        # connect defaults to ConnectStrategy() when none is provided
        cfg = databricks_config()
        assert cfg._credentials_strategy.auth_type() == ConnectStrategy().auth_type()



from datetime import datetime

import pytest
import responses

from posit.connect import Client
from posit.connect.external.aws import (
    Credentials,
    _decode_access_token,
    get_content_credentials,
    get_credentials,
)

aws_creds = Credentials(
    aws_access_key_id="abc123",
    aws_secret_access_key="def456",
    aws_session_token="ghi789",
    expiration=datetime(2025, 1, 1, 0, 0, 0, 0),
)

encoded_aws_creds = "eyJhY2Nlc3NLZXlJZCI6ICJhYmMxMjMiLCAic2VjcmV0QWNjZXNzS2V5IjogImRlZjQ1NiIsICJzZXNzaW9uVG9rZW4iOiAiZ2hpNzg5IiwgImV4cGlyYXRpb24iOiAiMjAyNS0wMS0wMVQwMDowMDowMFoifQ=="


class TestAWS:
    @responses.activate
    def test_get_aws_credentials(self):
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:user-session-token",
                        "subject_token": "cit",
                        "requested_token_type": "urn:ietf:params:aws:token-type:credentials",
                    }
                )
            ],
            json={
                "access_token": encoded_aws_creds,
                "issued_token_type": "urn:ietf:params:aws:token-type:credentials",
                "token_type": "aws_credentials",
            },
        )

        c = Client(api_key="12345", url="https://connect.example/")
        c._ctx.version = None
        response = get_credentials(c, "cit")

        assert response == aws_creds

    @responses.activate
    def test_get_aws_credentials_no_token(self):
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:user-session-token",
                        "subject_token": "cit",
                        "requested_token_type": "urn:ietf:params:aws:token-type:credentials",
                    }
                )
            ],
            json={},
        )

        c = Client(api_key="12345", url="https://connect.example/")
        c._ctx.version = None

        with pytest.raises(ValueError) as e:
            get_credentials(c, "cit")

        assert e.match("No access token found in credentials")

    @responses.activate
    def test_get_aws_content_credentials(self):
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:content-session-token",
                        "subject_token": "cit",
                        "requested_token_type": "urn:ietf:params:aws:token-type:credentials",
                    }
                )
            ],
            json={
                "access_token": encoded_aws_creds,
                "issued_token_type": "urn:ietf:params:aws:token-type:credentials",
                "token_type": "aws_credentials",
            },
        )

        c = Client(api_key="12345", url="https://connect.example/")
        c._ctx.version = None
        response = get_content_credentials(c, "cit")

        assert response == aws_creds

    @responses.activate
    def test_get_aws_content_credentials_no_token(self):
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:content-session-token",
                        "subject_token": "cit",
                        "requested_token_type": "urn:ietf:params:aws:token-type:credentials",
                    }
                )
            ],
            json={},
        )

        c = Client(api_key="12345", url="https://connect.example/")
        c._ctx.version = None

        with pytest.raises(ValueError) as e:
            get_content_credentials(c, "cit")

        assert e.match("No access token found in credentials")

    def test_decode_access_token(self):
        decoded_creds = _decode_access_token(encoded_aws_creds)
        assert decoded_creds == aws_creds
